#!/usr/bin/python-init -Ot
#
# Copyright (C) 2001,2002,2003,2004,2005,2006,2007,2008,2009,2012 Andreas Lang-Nevyjel
#
# Send feedback to: <lang-nevyjel@init.at>
#
# this file belongs to python-modules
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
""" interface to MySQL """

from pkg_resources import require
try:
    require("MySQL_python")
except:
    raise ImportError
try:
    require("PsycoPG2")
except:
    psycopg2 = None
else:
    import psycopg2
    import psycopg2.extensions
    import psycopg2.extras
import MySQLdb
import MySQLdb.cursors
import MySQLdb.converters
import configfile
import logging_tools
import io_stream_helper
import sys
import os
import os.path
import time
import threading
import pprint
import contextlib

    
class db_validate(object):
    def __init__(self, log, dc, **args):
        self.__dc = dc
        self.__act_db = args.get("database", "")
        self.__check_type = args.get("check_type", "MEDIUM")
        self.__repair_type = args.get("repair_type", "EXTENDED")
        self.__like_pattern = args.get("like_pattern", "")
        self._log = log
        self._fetch_table_names()
        self._check_tables()
    def log(self, what, log_level=logging_tools.LOG_LEVEL_OK):
        self._log(what, log_level)
    def _fetch_table_names(self):
        self.__dc.execute("SHOW TABLES%s%s" % (" FROM %s" % (self.__act_db) if self.__act_db else "",
                                               " LIKE '%s'" % (self.__like_pattern) if self.__like_pattern else ""))
        self.__table_names = [db_rec.values()[0] for db_rec in self.__dc.fetchall()]
        if self.__act_db:
            self.__table_names = ["%s.%s" % (self.__act_db, t_name) for t_name in self.__table_names]
        self.__damaged_tables = []
    def _check_tables(self):
        self.__dc.execute("CHECK TABLE %s %s" % (", ".join(self.__table_names),
                                                 self.__check_type))
        for db_rec in self.__dc.fetchall():
            if db_rec["Msg_type"].lower() == "error":
                self.__damaged_tables.append(db_rec["Table"])
        if self.__damaged_tables:
            self.log("found %s: %s" % (logging_tools.get_plural("damaged table", len(self.__damaged_tables)),
                                       ", ".join(sorted(self.__damaged_tables))),
                     logging_tools.LOG_LEVEL_ERROR)
        else:
            self.log("checked %s, all ok" % (logging_tools.get_plural("table", len(self.__table_names))))
    def get_damaged_tables(self):
        return self.__damaged_tables
    def repair_tables(self):
        for dt_name in self.__damaged_tables:
            s_time = time.time()
            self.__dc.execute("REPAIR TABLE %s %s" % (dt_name,
                                                      self.__repair_type))
            e_time = time.time()
            self.log("repaired table '%s' in %s (repair_type is %s)" % (dt_name,
                                                                        logging_tools.get_diff_time_str(e_time - s_time),
                                                                        self.__repair_type))
class mysql_dbcon_inst(MySQLdb.cursors.DictCursor):
    def __init__(self, container, idx, access_dict, **args):
        self.__idx = idx
        self.__container = container
        self.__access_dict = dict([(key, value) for key, value in access_dict.iteritems()])
        self.set_parameters(**args)
        self.__db = None
        self.__opt_dict = args.get("opt_dict", {})
        self.open_connection()
        self.init_counters()
        self.set_req_thread()
    def set_req_thread(self, rt=None):
        self.__req_thread = rt or threading.currentThread().getName()
    def get_req_thread(self):
        return self.__req_thread
    def set_parameters(self, **args):
        self.set_logging(args.get("with_logging", False))
        self.__catch_errors = args.get("catch_errors", False)
        if "database_name" in args:
            self.__access_dict["MYSQL_DATABASE"] = args["database_name"]
    def init_counters(self):
        self.__execute_counter, self.__fetchall_counter, self.__fetchone_counter = (0, 0, 0)
        self.__execute_error,   self.__fetchall_errors,  self.__fetchone_errors  = (0, 0, 0)
    def get_db(self):
        return self.__db
    def get_idx(self):
        return self.__idx
    def release(self):
        # release actual connection
        self.__container.release_connection(self)
    def set_logging(self, lf):
        self.__log_enabled = lf
    def enable_logging(self, lf):
        # deprecated, for compatibility
        self.__log_enabled = lf
    def close(self):
        self.close_connection()
        self.__container = None
        self.__idx = None
    def open_connection(self):
        if self.__db:
            self.__db.close()
        try:
            self.__db = MySQLdb.connect(self.__access_dict["MYSQL_HOST"],
                                        self.__access_dict["MYSQL_USER"],
                                        self.__access_dict["MYSQL_PASSWD"],
                                        self.__access_dict["MYSQL_DATABASE"])
        except MySQLdb.OperationalError, what:
            self.err_it(("db", "MySQL Operational error in __init__() : %s" % (what), 0.))
            self.__db = None
            if not self.__catch_errors:
                raise MySQLdb.OperationalError, what
        else:
            MySQLdb.cursors.DictCursor.__init__(self, self.__db)
            self.__db.set_character_set(self.__opt_dict.get("character_set", "utf8"))
            self.__db.autocommit(self.__opt_dict.get("auto_commit", True))
    def connection_valid(self):
        return self.__db and True or False
    def log_it(self, what):
        self.__container.log(what)
    def err_it(self, what):
        self.__container.log_err(what)
    def __del__(self):
        self.close_connection()
    def close_connection(self):
        if self.__db:
            self.__db.close()
        self.__db = None
    def get_exec_counter(self):
        return self.__execute_counter
    def get_total_counter(self):
        return self.__execute_counter + self.__fetchall_counter + self.__fetchone_counter
    def get_error_counter(self):
        return self.__execute_error + self.__fetchall_errors + self.__fetchone_errors
    def execute(self, query_str, args=None):
        if not self.__db:
            self.open_connection()
        self.__execute_counter += 1
        start_time = time.time()
        done, err_str = (False, "")
        retry_it, retry_num = (True, 0)
        while retry_it and retry_num < 4:
            retry_it = False
            try:
                MySQLdb.cursors.DictCursor.execute(self, query_str, args)
            except MySQLdb.OperationalError, (num, what):
                if num == 2006:
                    # try to reopen connection
                    self.err_it(("dc",
                                 "reopen DB-connection %d, pid %d, query_str is %s, %s" % (self.__idx,
                                                                                           os.getpid(),
                                                                                           query_str,
                                                                                           logging_tools.get_plural("argument", args and len(args) or 0)),
                                 0.))
                    self.open_connection()
                    retry_it = True
                    retry_num += 1
                elif num in [1053, 2013, 2002]:
                    # 1053: Server shutdown in progress
                    # 2013: Lost connection to MySQL server during query
                    # 2002: Can't connect to local MySQL server through socket '/var/lib/mysql/mysql.sock' (2)
                    time.sleep(60)
                    # try to reopen connection
                    self.err_it(("dc",
                                 "reopen DB-connection %d, pid %d, query_str is %s, %s" % (self.__idx,
                                                                                           os.getpid(),
                                                                                           query_str,
                                                                                           logging_tools.get_plural("argument", args and len(args) or 0)),
                                 0.))
                    self.open_connection()
                    retry_it = True
                    retry_num += 1
                else:
                    err_str = "MySQL Operational error %d in execute() : %s (%s)" % (num, what, query_str)
            except MySQLdb.ProgrammingError, (num, what):
                err_str = "MySQL Programming error %d in execute() : %s (%s)" % (num, what, query_str)
            except MySQLdb.IntegrityError, (num, what):
                err_str = "MySQL Integrity error %d in execute() : %s (%s)" % (num, what, query_str)
            except MySQLdb.InternalError, (num, what):
                err_str = "MySQL Internal error %d in execute() : %s (%s)" % (num, what, query_str)
            else:
                done = True
        if err_str:
            print "***", err_str
            raise SyntaxError
        end_time = time.time()
        if err_str:
            err_h = io_stream_helper.io_stream("/var/lib/logging-server/py_err")
            err_h.write("\n".join([query_str, str(args), err_str, str(sys.exc_info()[0]), str(sys.exc_info()[1])]))
            err_h.close()
            del err_h
            self.__execute_error += 1
            self.err_it(("dc", "(%d) %s" % (self.__idx, err_str), end_time - start_time))
        if self.__log_enabled:
            if args:
                log_str = query_str % args
            else:
                log_str = query_str
            self.log_it(("dc", "(%d) %s" % (self.__idx, log_str), end_time - start_time))
        return done
    def fetchall(self):
        self.__fetchall_counter += 1
        try:
            ret = MySQLdb.cursors.DictCursor.fetchall(self)
        except MySQLdb.ProgrammingError, what:
            self.__fetchall_errors += 1
            self.err_it(("dc", "MySQL Programming error in fetchall() for %d: %s" % (self.__idx,
                                                                                     what), 0.))
            ret = None
        except MySQLdb.OperationalError, what:
            self.__fetchall_errors += 1
            self.err_it(("dc", "MySQL Operational error in fetchall() for %d: %s" % (self.__idx,
                                                                                     what), 0.))
            ret = None
        return ret
    def fetchone(self):
        self.__fetchone_counter += 1
        try:
            ret = MySQLdb.cursors.DictCursor.fetchone(self)
        except MySQLdb.ProgrammingError, what:
            self.__fetchone_errors += 1
            self.err_it(("dc", "MySQL Programming error in fetchone() for %d: %s" % (self.__idx,
                                                                                     what), 0.))
            ret = None
        except MySQLdb.OperationalError, what:
            self.__fetchone_errors += 1
            self.err_it(("dc", "MySQL Operational error in fetchone() for %d: %s" % (self.__idx,
                                                                                     what), 0.))
            ret = None
        return ret
    def insert_id(self):
        try:
            ret = self.lastrowid
        except MySQLdb.ProgrammingError, what:
            self.err_it(("dc", "MySQL Programming error in insert_id() for %d: %s" % (self.__idx,
                                                                                      what), 0.))
            ret = None
        except MySQLdb.OperationalError, what:
            self.err_it(("dc", "MySQL Operational error in insert_id() for %d: %s" % (self.__idx,
                                                                                      what), 0.))
            ret = None
        return ret

class postgresql_dbcon_inst(psycopg2.extras.RealDictCursor):
    def __init__(self, container, idx, access_dict, **args):
        self.__idx = idx
        self.__container = container
        self.__access_dict = dict([(key, value) for key, value in access_dict.iteritems()])
        self.set_parameters(**args)
        self.__db = None
        self.open_connection()
        self.init_counters()
        self.set_req_thread()
    def set_req_thread(self, rt=None):
        self.__req_thread = rt or threading.currentThread().getName()
    def get_req_thread(self):
        return self.__req_thread
    def set_parameters(self, **args):
        self.set_logging(args.get("with_logging", False))
        self.__catch_errors = args.get("catch_errors", False)
        if "database_name" in args:
            self.__access_dict["MYSQL_DATABASE"] = args["database_name"]
    def init_counters(self):
        self.__execute_counter, self.__fetchall_counter, self.__fetchone_counter = (0, 0, 0)
        self.__execute_error,   self.__fetchall_errors,  self.__fetchone_errors  = (0, 0, 0)
    def get_db(self):
        return self.__db
    def get_idx(self):
        return self.__idx
    def release(self):
        # release actual connection
        self.__container.release_connection(self)
    def set_logging(self, lf):
        self.__log_enabled = lf
    def enable_logging(self, lf):
        # deprecated, for compatibility
        self.__log_enabled = lf
    def close(self):
        self.close_connection()
        self.__container = None
        self.__idx = None
    def open_connection(self):
        if self.__db:
            self.__db.close()
        try:
            self.__db = psycopg2.extras.RealDictConnection("dbname='%s' user='%s' host='%s' password='%s'" % (self.__access_dict["MYSQL_DATABASE"],
                                                                                                              self.__access_dict["MYSQL_USER"],
                                                                                                              self.__access_dict["MYSQL_HOST"],
                                                                                                              self.__access_dict["MYSQL_PASSWD"]))
        except psycopg2.OperationalError, what:
            self.err_it(("db", "PostGRESQL Operational error in __init__() : %s" % (what), 0.))
            self.__db = None
            if not self.__catch_errors:
                raise psycopg2.OperationalError, what
        else:
            psycopg2.extras.RealDictCursor.__init__(self, self.__db)
    def connection_valid(self):
        return self.__db and True or False
    def log_it(self, what):
        self.__container.log(what)
    def err_it(self, what):
        self.__container.log_err(what)
    def __del__(self):
        self.close_connection()
    def close_connection(self):
        if self.__db:
            self.__db.close()
        self.__db = None
    def get_exec_counter(self):
        return self.__execute_counter
    def get_total_counter(self):
        return self.__execute_counter + self.__fetchall_counter + self.__fetchone_counter
    def get_error_counter(self):
        return self.__execute_error + self.__fetchall_errors + self.__fetchone_errors
    def execute(self, query_str, args=None):
        if not self.__db:
            self.open_connection()
        self.__execute_counter += 1
        start_time = time.time()
        done, err_str = (False, "")
        retry_it, retry_num = (True, 0)
        while retry_it and retry_num < 2:
            retry_it = False
            try:
                psycopg2.extras.RealDictCursor.execute(self, query_str, args)
            except psycopg2.OperationalError, (num, what):
                if num == 2006:
                    # try to reopen connection
                    self.err_it(("dc",
                                 "reopen DB-connection %d, pid %d, query_str is %s, %s" % (self.__idx,
                                                                                           os.getpid(),
                                                                                           query_str,
                                                                                           logging_tools.get_plural("argument", args and len(args) or 0)),
                                 0.))
                    self.open_connection()
                    retry_it = True
                    retry_num += 1
                else:
                    err_str = "PostGRESQL Operational error %d in execute() : %s (%s)" % (num, what, query_str)
            except psycopg2.ProgrammingError, what:
                err_str = "PostGRESQL Programming error in execute() : %s (%s)" % (what, query_str)
            except psycopg2.IntegrityError, (num, what):
                err_str = "PostGRESQL Integrity error %d in execute() : %s (%s)" % (num, what, query_str)
            except psycopg2.InternalError, (num, what):
                err_str = "PostGRESQL Internal error %d in execute() : %s (%s)" % (num, what, query_str)
            else:
                done = True
        end_time = time.time()
        if err_str:
            err_h = io_stream_helper.io_stream("/var/lib/logging-server/py_err")
            err_h.write("\n".join([query_str, str(args), err_str, str(sys.exc_info()[0]), str(sys.exc_info()[1])]))
            err_h.close()
            del err_h
            self.__execute_error += 1
            self.err_it(("dc", "(%d) %s" % (self.__idx, err_str), end_time - start_time))
        if self.__log_enabled:
            if args:
                log_str = query_str % args
            else:
                log_str = query_str
            self.log_it(("dc", "(%d) %s" % (self.__idx, log_str), end_time - start_time))
        return done
    def fetchall(self):
        self.__fetchall_counter += 1
        try:
            ret = psycopg2.extras.RealDictCursor.fetchall(self)
        except psycopg2.ProgrammingError, what:
            self.__fetchall_errors += 1
            self.err_it(("dc", "PostGRESQL Programming error in fetchall() for %d: %s" % (self.__idx,
                                                                                     what), 0.))
            ret = None
        except psycopg2.OperationalError, what:
            self.__fetchall_errors += 1
            self.err_it(("dc", "PostGRESQL Operational error in fetchall() for %d: %s" % (self.__idx,
                                                                                     what), 0.))
            ret = None
        return ret
    def fetchone(self):
        self.__fetchone_counter += 1
        try:
            ret = psycopg2.extras.RealDictCursor.fetchone(self)
        except psycopg2.ProgrammingError, what:
            self.__fetchone_errors += 1
            self.err_it(("dc", "PostGRESQL Programming error in fetchone() for %d: %s" % (self.__idx,
                                                                                     what), 0.))
            ret = None
        except psycopg2.OperationalError, what:
            self.__fetchone_errors += 1
            self.err_it(("dc", "PostGRESQL Operational error in fetchone() for %d: %s" % (self.__idx,
                                                                                     what), 0.))
            ret = None
        return ret
    def commit(self):
        self.__db.commit()
    def abort(self):
        self.__db.abort()
    def insert_id(self):
        try:
            ret = self.lastrowid
        except psycopg2.ProgrammingError, what:
            self.err_it(("dc", "PostGRESQL Programming error in insert_id() for %d: %s" % (self.__idx,
                                                                                      what), 0.))
            ret = None
        except psycopg2.OperationalError, what:
            self.err_it(("dc", "PostGRESQL Operational error in insert_id() for %d: %s" % (self.__idx,
                                                                                      what), 0.))
            ret = None
        return ret

class dbcon_container(object):
    # container for database connectors
    def __init__(self, **args):
        self.__with_logging = os.path.isfile("/etc/.mysql_debug") or args.get("with_logging", False)
        self.__catch_errors = args.get("catch_errors", False)
        self.__opt_dict = {"auto_commit"   : args.get("auto_commit", True),
                           "character_set" : args.get("character_set", "utf8")}
        if args.get("thread_safe", True):
            self.__lock = threading.Lock()
        else:
            self.__lock = None
        # connection pool
        self.__con_pool, self.__con_lut = ({}, {})
        self.set_add_data(args.get("add_data", None))
        self.set_log_instance(log_handler=args.get("log_handler", None),
                              err_handler=args.get("err_handler", None))
        self.__has_access_file = args.get("has_access_file", True)
        if self.__has_access_file:
            self._init_access_from_file(cf_name=args.get("cf_name", "/etc/sysconfig/cluster/db_access"))
        else:
            self._init_access_from_arg(args["access_dict"])
    def lock(self):
        if self.__lock:
            self.__lock.acquire()
    def release(self):
        if self.__lock:
            self.__lock.release()
    def _init_access_from_arg(self, cf_dict):
        key = "default"
        self.__access_dict, self.__default_access = ({}, key)
        self.__access_dict[key] = {"access_data_ok" : True,
                                   "access_data"    : cf_dict}
        self.__con_pool[key] = {"next_idx"       : 1,
                                "allocated_cons" : {},
                                "free_cons"      : {}}
        self.check_default_access()
    def _init_access_from_file(self, cf_name="/etc/sysconfig/cluster/db_access"):
        self.__access_dict, self.__default_access = ({}, None)
        if os.path.isfile(cf_name):
            csuc, cfile = configfile.readconfig(cf_name, 1)
            if csuc:
                read_errors, any_ok = ([], False)
                for key, value in cfile.iteritems():
                    if key == "default_access":
                        self.__default_access = value
                    else:
                        k_parts = [line.strip() for line in value.split(";")]
                        # k_parts: [file_name, {add_stuff}]
                        self.__access_dict[key] = {"file_name"      : k_parts[0],
                                                   "access_data_ok" : False,
                                                   "add_data"       : len(k_parts) > 1 and k_parts[1:] or []}
                        # try to populate access_data
                        try:
                            ac_ok, ac_data = configfile.readconfig(self.__access_dict[key]["file_name"], 1)
                        except:
                            pass
                        else:
                            if ac_ok:
                                self.__access_dict[key]["access_data_ok"] = True
                                self.__access_dict[key]["access_data"] = dict([(i_key, i_value) for i_key, i_value in ac_data.iteritems() if [True for m_key in ["mysql", "is"] if i_key.lower().startswith(m_key)]])
                                # init connection_pool for type
                                self.__con_pool[key] = {"next_idx"       : 1,
                                                        "allocated_cons" : {},
                                                        "free_cons"      : {}}
                                any_ok = True
                            else:
                                read_errors.append(self.__access_dict[key]["file_name"])
                if not any_ok:
                    for file_name in read_errors:
                        self.log_err(("co", "cannot read data from %s" % (file_name), 0.))
        else:
            self.log_err(("co", "cannot find access_file %s" % (cf_name), 0.))
        if self.__with_logging:
            self.log(("co", "Initiated access_dict, %s of %d" % (logging_tools.get_plural("valid access type", len([True for key, value in self.__access_dict.iteritems() if value["access_data_ok"]])),
                                                                 len(self.__access_dict.keys())), 0.))
        self.check_default_access()
##         pprint.pprint(self.__access_dict)
##         print self.__default_access
    def check_default_access(self):
        if not self.__default_access in self.__access_dict:
            self.__default_access = None
        if not self.__default_access and self.__access_dict:
            self.__default_access = self.__access_dict.keys()[0]
        if self.__with_logging:
            if self.__default_access:
                self.log(("co", "set %s as default_access" % (self.__default_access), 0.))
            else:
                self.log_err(("co", "unable to set default_access", 0.))
    def get_access_dict(self, con_type=None):
        if not con_type:
            con_type = self.__default_access
        return self.__access_dict[con_type]
    def set_add_data(self, add_data=None):
        self.__add_data = add_data
    def set_log_instance(self, log_handler=None, err_handler=None):
        self.__lh, self.__eh = (log_handler, err_handler)
    def log(self, what):
        if self.__lh:
            self.__lh(what, self.__add_data)
        else:
            frm, what, tc = what
            logging_tools.my_syslog("pid %d: %s: %s (%.2f secs)" % (os.getpid(), frm, what, tc))
    def log_err(self, what):
        if self.__eh:
            self.__eh(what, self.__add_data)
        else:
            frm, what, tc = what
            logging_tools.my_syslog("pid %d:E: %s: %s (%.2f secs)" % (os.getpid(), frm, what, tc))
    def get_connection(self, con_type=None, **args):
        if con_type is None:
            con_type = self.__default_access
        with_logging = args.get("with_logging", self.__with_logging)
        catch_errors = args.get("catch_errors", self.__catch_errors)
        if con_type in self.__access_dict:
            if self.__access_dict[con_type]["access_data_ok"]:
                database_name = args.get("database_name", self.__access_dict[con_type]["access_data"]["MYSQL_DATABASE"])
                self.lock()
                act_cps = self.__con_pool[con_type]
                if act_cps["free_cons"]:
                    # there are free connections, reuse them
                    new_idx = act_cps["free_cons"].keys()[0]
                    new_con = act_cps["free_cons"][new_idx]
                    # set with_logging and catch_errors
                    new_con.set_parameters(with_logging=with_logging, catch_errors=catch_errors, database_name=database_name)
                    # reinit counters
                    new_con.init_counters()
                    new_con.open_connection()
                    act_cps["allocated_cons"][new_idx] = new_con
                    del act_cps["free_cons"][new_idx]
                    if self.__with_logging:
                        self.log(("co", "moving connection idx %d from free to allocated list (type %s)" % (new_idx, con_type), 0.))
                    new_con.set_req_thread()
                else:
                    if self.__access_dict[con_type]["access_data"].get("IS_POSTGRESQL", False):
                        new_con = postgresql_dbcon_inst(self,
                                                   act_cps["next_idx"],
                                                   self.__access_dict[con_type]["access_data"],
                                                   with_logging=with_logging,
                                                   catch_errors=catch_errors,
                                                   database_name=database_name)
                    else:
                        new_con = mysql_dbcon_inst(self,
                                                   act_cps["next_idx"],
                                                   self.__access_dict[con_type]["access_data"],
                                                   with_logging=with_logging,
                                                   catch_errors=catch_errors,
                                                   database_name=database_name,
                                                   opt_dict=self.__opt_dict)
                    act_cps["allocated_cons"][new_con.get_idx()] = new_con
                    self.__con_lut[new_con.get_idx()] = con_type
                    act_cps["next_idx"] += 1
                    if self.__with_logging:
                        self.log(("co", "allocated new connection idx %d (type %s)" % (new_con.get_idx(), con_type), 0))
                if self.__with_logging:
                    self.log_connection_stats()
                self.release()
            else:
                self.log_err(("co", "access_data invalid for access_type '%s'" % (con_type), 0.))
                raise ValueError, "access_data for access_type '%s' invalid" % (con_type)
        else:
            self.log_err(("co", "unknown access_type '%s' (valid: %s)" % (con_type, ", ".join(self.__access_dict.keys())), 0.))
            raise KeyError, "access_type %s not known" % (con_type)
        return new_con
    def log_connection_stats(self):
        v_cons = sorted(self.__con_pool.keys())
        log_str = ", ".join(["%s: %d/%d" % (v_con,
                                            len(self.__con_pool[v_con]["allocated_cons"].keys()),
                                            len(self.__con_pool[v_con]["free_cons"].keys())) for v_con in v_cons])
        self.log(("co", log_str, 0.))
    def release_connection(self, con):
        self.lock()
        con_type = self.__con_lut[con.get_idx()]
        if self.__with_logging:
            self.log(("co",
                      "freeing connection idx %d (type %s)" % (con.get_idx(),
                                                               con_type),
                      0.))
        act_cps = self.__con_pool[con_type]
        del act_cps["allocated_cons"][con.get_idx()]
        act_cps["free_cons"][con.get_idx()] = con
        # close db-connection
        con.close_connection()
        if self.__with_logging:
            self.log_connection_stats()
        self.release()
    def close(self):
        v_cons = sorted(self.__con_pool.keys())
        any_used = sum([self.__con_pool[v_con]["allocated_cons"].keys() for v_con in v_cons], [])
        if any_used:
            self.log_err(("co",
                          "%s found: %s" % (logging_tools.get_plural("unreleased connection", len(any_used)),
                                            ", ".join(["%d: %s" % (x, self.__con_lut[x]) for x in any_used])),
                          0))
            raise ValueError, "some connections (%d: %s) unreleased" % (len(any_used),
                                                                        ", ".join(["%d (type %s, thread %s)" % (x,
                                                                                                                self.__con_lut[x],
                                                                                                                self.__con_pool[self.__con_lut[x]]["allocated_cons"][x].get_req_thread()) for x in any_used]))
        for v_con in v_cons:
            for con in self.__con_pool[v_con]["free_cons"].values():
                con.close()
    @contextlib.contextmanager
    def db_transaction(self, con_type=None, **args):
        cursor = self.get_connection(con_type, **args)
        try:
            yield cursor
        except:
            #connection.rollback()
            # re-raise exception
            raise
        else:
            #connection.commit()
            pass
        finally:
            cursor.release()
    def __del__(self):
        pass

def device_log_entry(dc, device, log_source, user, log_status, text):
    sql_str, sql_tuple = get_device_log_entry_part(device, log_source, user, log_status, text)
    dc.execute("INSERT INTO devicelog VALUES(%s)" % (sql_str), sql_tuple)
    return dc.insert_id()

def get_device_log_entry_part(device, log_source, user, log_status, text):
    return "0, %s, %s, %s, %s, %s, null", (device, log_source, user, log_status, text)

def ext_device_log_entry(dc, devlog, log_source, user, users, subject, descr):
    dc.execute("INSERT INTO extended_log VALUES(0, %s, %s, %s, %s, %s, %s, null)", (devlog,
                                                                                    log_source,
                                                                                    user,
                                                                                    users,
                                                                                    subject,
                                                                                    descr))
    return dc.insert_id()

if __name__ == "__main__":
    print "Loadable module, exiting ..."
    sys.exit(0)
    
